#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_actor_runnable import IActorRunnable
from fabric.actor.core.apis.i_reservation import ReservationCategory
from fabric.actor.core.common.constants import Constants, ErrorCodes
from fabric.actor.core.core.authority_policy import AuthorityPolicy
from fabric.actor.core.kernel.broker_reservation_factory import BrokerReservationFactory
from fabric.actor.core.kernel.reservation_factory import ReservationFactory
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.kernel.slice import SliceTypes
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.manage.actor_management_object import ActorManagementObject
from fabric.actor.core.manage.converter import Converter
from fabric.actor.core.manage.management_object import ManagementObject
from fabric.actor.core.manage.messages.result_client_mng import ResultClientMng
from fabric.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric.message_bus.messages.result_string_avro import ResultStringAvro
from fabric.actor.core.proxies.kafka.translate import Translate
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.resource_data import ResourceData
from fabric.message_bus.messages.result_avro import ResultAvro
from fabric.message_bus.messages.result_slice_avro import ResultSliceAvro
from fim.graph.abc_property_graph import ABCPropertyGraph

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_server_actor import IServerActor
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.manage.messages.client_mng import ClientMng
    from fabric.message_bus.messages.slice_avro import SliceAvro


class ServerActorManagementObject(ActorManagementObject):
    def __init__(self, *, sa: IServerActor = None):
        super().__init__(actor=sa)

    def save(self) -> dict:
        properties = super().save()
        properties[Constants.property_class_name] = ServerActorManagementObject.__name__
        properties[Constants.property_module_name] = ServerActorManagementObject.__module__
        return properties

    def get_reservations_by_category(self, *, caller: AuthToken, category: ReservationCategory, slice_id: ID = None,
                                     id_token: str = None) -> ResultReservationAvro:
        result = ResultReservationAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result
        try:
            res_list = None
            try:
                if category == ReservationCategory.Client:
                    if slice_id is None:
                        res_list = self.db.get_client_reservations()
                    else:
                        res_list = self.db.get_client_reservations_by_slice_id(slice_id=slice_id)
                elif category == ReservationCategory.Broker:
                    res_list = self.db.get_broker_reservations()
                elif category == ReservationCategory.Inventory:
                    if slice_id is None:
                        res_list = self.db.get_holdings()
                    else:
                        res_list = self.db.get_holdings_by_slice_id(slice_id=slice_id)
            except Exception as e:
                self.logger.error("do_get_reservations:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.name)
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                return result

            if res_list is not None:
                result.result = []
                for r in res_list:
                    slice_obj = self._get_slice_by_id(id=r['slc_id'])
                    rsv_obj = ReservationFactory.create_instance(properties=r, actor=self.actor,
                                                                 slice_obj=slice_obj,
                                                                 logger=self.actor.get_logger())
                    if rsv_obj is not None:
                        rr = Converter.fill_reservation(reservation=rsv_obj, full=False)
                        result.result.append(rr)
        except Exception as e:
            self.logger.error("do_get_reservations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_slices_by_slice_type(self, *, caller: AuthToken, slice_type: SliceTypes,
                                 id_token: str = None) -> ResultSliceAvro:
        result = ResultSliceAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            slc_list = None

            try:
                if slice_type == SliceTypes.ClientSlice:
                    slc_list = self.db.get_client_slices()
                elif slice_type == SliceTypes.InventorySlice:
                    slc_list = self.db.get_inventory_slices()

            except Exception as e:
                self.logger.error("get_slices_by_slice_type:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.name)
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                return result

            if slc_list is not None:
                result.result = Translate.fill_slices(slice_list=slc_list, full=True)

        except Exception as e:
            self.logger.error("get_slices_by_slice_type: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def add_client_slice(self, *, caller: AuthToken, slice_mng: SliceAvro) -> ResultSliceAvro:
        result = ResultSliceAvro()
        result.status = ResultAvro()

        if caller is None or slice_mng is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            owner_mng = slice_mng.get_owner()
            owner = None
            owner_is_ok = False

            if owner_mng is not None:
                owner = Translate.translate_auth_from_avro(auth_avro=owner_mng)
                if owner is not None and owner.get_name() is not None and owner.get_guid() is not None:
                    owner_is_ok = True

            if not owner_is_ok:
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.name)
                return result

            slice_obj = SliceFactory.create(slice_id=ID(), name=slice_mng.get_slice_name(),
                                            data=Converter.get_resource_data(slice_mng=slice_mng))
            slice_obj.set_description(description=slice_mng.get_description())
            slice_obj.set_inventory(value=False)

            assert owner is not None
            slice_obj.set_owner(owner=owner)

            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    try:
                        self.actor.register_slice(slice_object=slice_obj)
                    except Exception as e:
                        self.actor.get_plugin().release_slice(slice_obj=slice_obj)
                        raise e

                    return None

            self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))

            result.result = str(slice_obj.get_slice_id())

        except Exception as e:
            self.logger.error("addClientSlice: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def register_client(self, *, client: ClientMng, kafka_topic: str, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if client is None or kafka_topic is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            client_obj = Converter.fill_client(client_mng=client)
            client_obj.set_kafka_topic(kafka_topic=kafka_topic)

            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    self.actor.register_client(client=client_obj)
                    return None

            self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))

        except Exception as e:
            self.logger.error("register_client: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def get_clients(self, *, caller: AuthToken, guid: ID = None, id_token: str = None) -> ResultClientMng:
        result = ResultClientMng()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            cl_list = None
            if guid is None:
                cl_list = self.db.get_clients()
            else:
                cl = self.db.get_client(guid=guid)
                if cl is not None:
                    cl_list = [cl]
                else:
                    result.status.set_code(ErrorCodes.ErrorNoSuchActor.value)
                    result.status.set_message(ErrorCodes.ErrorNoSuchActor.name)

            if cl_list is not None:
                result.result = Converter.fill_clients(client_list=cl_list)
        except Exception as e:
            self.logger.error("get_clients: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

    def unregister_client(self, *, guid: ID, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()

        if guid is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    self.actor.unregister_client(guid=guid)
                    return None

            self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))

        except Exception as e:
            self.logger.error("unregister_client: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def advertise_resources(self, *, delegation: ABCPropertyGraph, client: AuthToken, caller: AuthToken) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()

        if client is None or delegation is None or caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    return self.actor.advertise(delegation=delegation, client=client)

            self.logger.debug("Executing advertise on actor {} {} ({})".format(self.actor.get_name(),
                                                                               self.actor.get_name(),
                                                                               self.actor.__class__.__name__))

            advertised = self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))
            result.result = str(advertised)
        except Exception as e:
            self.logger.error("advertise_resources: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.common.exceptions import EventException
from fabric.actor.core.container.aevent_subscription import AEventSubscription

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_event_handler import IEventHandler
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.apis.i_event import IEvent


class SynchronousEventSubscription(AEventSubscription):
    def __init__(self, *, handler: IEventHandler, token: AuthToken, filters: list):
        super().__init__(token=token, filters=filters)
        self.handler = handler

    def deliver_event(self, *, event: IEvent):
        if self.matches(event=event):
            self.handler.handle(event=event)

    def drain_events(self, *, timeout: int) -> list:
        raise EventException("drainEvents is not supported on synchronous subscriptions")

    def is_abandoned(self) -> bool:
        return False



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING, Dict

from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.kernel.reservation_states import ReservationStates
from fabric.actor.core.util.prop_list import PropList
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.apis.i_broker_policy import IBrokerPolicy
from fabric.actor.core.core.policy import Policy
from fabric.actor.core.kernel.resource_set import ResourceSet

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker import IBroker
    from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.delegation.resource_delegation import ResourceDelegation
    from fabric.actor.core.apis.i_server_reservation import IServerReservation
    from fabric.actor.core.util.id import ID


class BrokerPolicy(Policy, IBrokerPolicy):
    """
    Base implementation for Broker policy
    """
    def __init__(self, *, actor: IBroker):
        super().__init__(actor=actor)
        # If true, every allocated ticket will require administrative approval
        # before being sent back.
        self.required_approval = False
        # A list of reservations that require administrative approval. Note: this
        # field is private so that derived classes are forced to use the locked
        # methods to operate on it.
        self.for_approval = None
        # Initialization status.
        self.initialized = False

        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['clock']
        del state['initialized']

        del state['for_approval']
        del state['lock']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.actor = None
        self.clock = None
        self.initialized = False

        self.lock = threading.Lock()

    def add_for_approval(self, *, reservation: IBrokerReservation):
        """
        Adds the reservation to the approval list.

        @param reservation reservation
        """
        try:
            self.lock.acquire()
            self.for_approval.add(reservation=reservation)
        finally:
            self.lock.release()

    def allocate(self, *, cycle: int):
        return

    def approve(self, *, reservation: IBrokerReservation):
        """
        Approve a reservation. To be used by policies that require administrative
        intervention. Override to provide your own approval policy.

        @param reservation reservation
        """

    def bind(self, *, reservation: IBrokerReservation) -> bool:
        return False

    def bind_delegation(self, *, delegation: IDelegation) -> bool:
        return False

    def demand(self, *, reservation: IClientReservation):
        return

    def donate_reservation(self, *, reservation: IClientReservation):
        return

    def donate_delegation(self, *, delegation: IDelegation):
        return

    def extend_broker(self, *, reservation: IBrokerReservation) -> bool:
        return False

    def initialize(self):
        if not self.initialized:
            super().initialize()
            self.initialized = True

    def release_not_approved(self, *, reservation: IBrokerReservation):
        """
        Releases the resources for a reservation that was rejected by the
        administrator.

        @param reservation reservation
        """

    def remove_for_approval(self, *, reservation: IBrokerReservation):
        """
        Remove reservation from approval list
        """
        try:
            self.lock.acquire()
            self.for_approval.remove(reservation=reservation)
        finally:
            self.lock.release()

    def revisit(self, *, reservation: IReservation):
        if isinstance(reservation, IClientReservation) and reservation.get_state() == ReservationStates.Ticketed:
            self.donate_reservation(reservation=reservation)

    def revisit_delegation(self, *, delegation: IDelegation):
        if delegation.get_state() == ReservationStates.Ticketed:
            self.donate_delegation(delegation=delegation)

    def ticket_satisfies(self, *, requested_resources: ResourceSet, actual_resources: ResourceSet,
                         requested_term: Term, actual_term: Term):
        return

    def update_ticket_complete(self, *, reservation: IClientReservation):
        self.donate_reservation(reservation=reservation)

    def update_delegation_complete(self, *, delegation: IDelegation):
        self.donate_delegation(delegation=delegation)

    def extract(self, *, source: ResourceSet, delegation: ResourceDelegation) -> ResourceSet:
        """
        Creates a new resource set using the source and the specified delegation.

        @param source source
        @param delegation delegation
        @return returns ResourceSet
        @throws Exception in case of error
        """
        rd = ResourceData()
        temp = PropList.merge_properties(incoming=source.get_resource_properties(),
                                         outgoing=rd.get_resource_properties())
        rd.resource_properties = temp

        extracted = ResourceSet(units=delegation.get_units(), rtype=delegation.get_resource_type(), rdata=rd)

        ticket = source.get_resources()
        resource_ticket = source.get_resources().get_ticket()

        new_ticket = self.actor.get_plugin().get_ticket_factory().make_ticket(delegation=delegation,
                                                                              source=resource_ticket)

        cset = Ticket(plugin=ticket.get_plugin(), ticket=new_ticket, authority=ticket.get_authority())

        extracted.set_resources(cset=cset)
        return extracted

    def get_client_id(self, *, reservation: IServerReservation) -> ID:
        """
        Get Client Id
        """
        return reservation.get_client_auth_token().get_guid()

    @staticmethod
    def get_resource_pools_query() -> dict:
        """
        Return dictionary representing query
        """
        properties = {Constants.query_action: Constants.query_action_discover_pools}
        return properties

    @staticmethod
    def get_resource_pools(response: dict) -> Dict[ResourceType, ResourcePoolDescriptor]:
        """
        Get Resource Pools from Query response
        @param response query response
        """
        result = {}

        if Constants.pools_count in response:
            count = int(response[Constants.pools_count])
            for i in range(count):
                rd = ResourcePoolDescriptor()
                rd.reset(properties=response, prefix=Constants.pool_prefix + str(i) + ".")
                result[rd.get_resource_type()] = rd

        return result

    @staticmethod
    def get_query_action(properties: dict) -> str:
        if properties is None:
            return None
        return properties.get(Constants.query_action, None)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_event import IEvent
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reservation_state import ReservationState


class ReservationStateTransitionEvent(IEvent):
    """
    Represents State Transition Event
    """
    def __init__(self, *, reservation: IReservation, state: ReservationState):
        self.actor_id = reservation.get_actor().get_guid()
        self.rid = reservation.get_reservation_id()
        self.slice_id = reservation.get_slice_id()
        self.state = state
        self.reservation = reservation

    def get_actor_id(self) -> ID:
        """
        Get Actor id
        @return actor id
        """
        return self.actor_id

    def get_reservation_id(self) -> ID:
        """
        Get Reservation id
        @return reservation id
        """
        return self.rid

    def get_state(self) -> ReservationState:
        """
        Get reservation State
        @return state
        """
        return self.state

    def get_reservation(self) -> IReservation:
        """
        Get Reservation
        @return reservation
        """
        return self.reservation

    def get_properties(self) -> dict:
        """
        Get properties
        @return properties
        """
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from datetime import datetime
from typing import List

from fabric.orchestrator.core.slice_state_machine import SliceStateMachine, SliceState
from fabric.message_bus.messages.reservation_mng import ReservationMng
from fabric.message_bus.messages.ticket_reservation_avro import TicketReservationAvro

from fabric.orchestrator.core.request_workflow import RequestWorkflow
from fabric.orchestrator.core.reservation_converter import ReservationConverter
from fabric.actor.core.apis.i_mgmt_controller import IMgmtController
from fabric.actor.core.util.id import ID
from fabric.message_bus.messages.slice_avro import SliceAvro


class OrchestratorSlice:
    def __init__(self, *, controller: IMgmtController, slice_obj: SliceAvro, name: str, user_dn: str,
                 ssh_credentials: list, recover: bool):
        self.slice_lock = threading.Lock()
        self.state_machine = SliceStateMachine(slice_id=ID(uid=slice_obj.get_slice_id()), recover=recover)
        self.slice_urn = name
        self.user_dn = user_dn
        self.ssh_credentials = ssh_credentials
        self.computed_reservations = None
        self.slice_obj = slice_obj
        self.workflow = None
        self.reservation_converter = None
        self.first_delete_attempt = None
        self.close_executed = False
        self.global_assignment_cleared = False

        try:
            cloud_handler = None
            # TODO
            # initialize the cloud handler
            self.workflow = RequestWorkflow(cloud_handler=cloud_handler)
            self.reservation_converter = ReservationConverter(ssh_credentials=self.ssh_credentials,
                                                              controller=controller, controller_slice=self)
        except Exception as e:
            raise e

    def lock(self):
        self.slice_lock.acquire()

    def unlock(self):
        self.slice_lock.release()

    def remove_computed_reservations(self, *, rid: str):
        if self.computed_reservations is not None:
            reservation_to_remove = None
            for reservation in self.computed_reservations:
                if reservation.get_reservation_id() == rid:
                    reservation_to_remove = reservation
                    break
            if reservation_to_remove is not None:
                self.computed_reservations.remove(reservation_to_remove)

    def add_computed_reservations(self, *, reservation: TicketReservationAvro):
        if self.computed_reservations is None:
            self.computed_reservations = []
        self.computed_reservations.append(reservation)

    def set_computed_reservations(self, *, reservations: List[TicketReservationAvro]):
        self.computed_reservations = reservations

    def get_computed_reservations(self) -> List[TicketReservationAvro]:
        return self.computed_reservations

    def get_all_reservations(self, *, controller: IMgmtController) -> List[ReservationMng]:
        if controller is None:
            return None

        return controller.get_reservations()

    def get_units(self, *, controller: IMgmtController, rid: ID):
        try:
            controller.get_reservation_units(rid=rid)
        except Exception as e:
            raise e

    def get_reservation_states(self, *, controller: IMgmtController, rids: List[ID]):
        try:
            if rids is None:
                return None
            return controller.get_reservation_state_for_reservations(reservation_list=rids)
        except Exception as e:
            raise e

    def get_workflow(self) -> RequestWorkflow:
        return self.workflow

    def get_user_dn(self) -> str:
        return self.user_dn

    def match_user_dn(self, *, dn: str):
        if dn is None:
            return False
        return self.user_dn == dn

    def get_slice_urn(self) -> str:
        return self.slice_urn

    def get_slice_id(self) -> ID:
        return ID(uid=self.slice_obj.get_slice_id())

    def get_reservation_converter(self) -> ReservationConverter:
        return self.reservation_converter

    def close(self):
        if not self.global_assignment_cleared and \
                (self.state_machine.get_state() == SliceState.Dead or
                 self.state_machine.get_state() == SliceState.Closing):
            self.workflow.close()

        if self.computed_reservations is not None:
            self.computed_reservations.clear()
            self.computed_reservations = None

    def get_state_machine(self) -> SliceStateMachine:
        return self.state_machine

    def is_stable_ok(self) -> bool:
        slice_state = self.reevaluate()

        if slice_state == SliceState.StableOk:
            return True

        return False

    def is_stable_error(self) -> bool:
        slice_state = self.reevaluate()

        if slice_state == SliceState.StableError:
            return True

        return False

    def is_stable(self) -> bool:
        slice_state = self.reevaluate()

        if slice_state == SliceState.StableError or slice_state == SliceState.StableOk:
            return True

        return False

    def is_dead_or_closing(self) -> bool:
        slice_state = self.reevaluate()

        if slice_state == SliceState.Dead or slice_state == SliceState.Closing:
            return True

        return False

    def is_dead(self) -> bool:
        slice_state = self.reevaluate()

        if slice_state == SliceState.Dead:
            return True

        return False

    def all_failed(self):
        return self.state_machine.all_failed()

    def mark_delete_attempt(self):
        if self.first_delete_attempt is None:
            self.first_delete_attempt = datetime.utcnow()

    def get_delete_attempt(self) -> datetime:
        return self.first_delete_attempt

    def reevaluate(self) -> SliceState:
        return self.state_machine.transition_slice(operation=SliceStateMachine.REEVALUATE)

    def recover(self):
        # TODO
        return

    def add_recover_reservation(self, *, r: ReservationMng):
        # TODO
        return

    def get_requested_entities(self) -> dict:
        ticketed_requested_entities = {}
        if self.get_computed_reservations() is not None:
            for reservation in self.get_computed_reservations():
                ticketed_requested_entities[reservation.get_reservation_id()] = reservation

        return ticketed_requested_entities

    def get_computed_reservation_summary(self) -> str:
        result = ""

        if self.get_computed_reservations() is not None:
            for reservation in self.get_computed_reservations():
                result += "[ Slice UID: {} | Reservation UID: {} | Resource Type: {} | Resource Units: {} ]\n".format(
                    reservation.get_slice_id(), reservation.get_reservation_id(), reservation.get_resource_type(),
                    reservation.get_units())
        else:
            result += "No new reservations were computed\n"

        if len(result) == 0:
            result += "No result available"

        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_authority import IAuthority
from fabric.actor.core.apis.i_broker import IBroker
from fabric.actor.core.proxies.local.local_authority import LocalAuthority
from fabric.actor.core.proxies.local.local_broker import LocalBroker
from fabric.actor.core.proxies.local.local_return import LocalReturn
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.apis.i_proxy_factory import IProxyFactory

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor_identity import IActorIdentity
    from fabric.actor.core.proxies.actor_location import ActorLocation
    from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy
    from fabric.actor.core.apis.i_proxy import IProxy


class LocalProxyFactory(IProxyFactory):
    def new_callback(self, *, identity: IActorIdentity, location: ActorLocation) -> ICallbackProxy:
        actor = ActorRegistrySingleton.get().get_actor(actor_name_or_guid=identity.get_name())
        if actor is not None:
            return LocalReturn(actor=actor)
        return None

    def new_proxy(self, *, identity: IActorIdentity, location: ActorLocation, proxy_type: str = None) -> IProxy:
        actor = ActorRegistrySingleton.get().get_actor(actor_name_or_guid=identity.get_name())
        if actor is not None:
            if isinstance(actor, IAuthority):
                return LocalAuthority(actor=actor)
            elif isinstance(actor, IBroker):
                return LocalBroker(actor=actor)
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.apis.i_event_filter import IEventFilter
    from fabric.actor.core.apis.i_event import IEvent


class AEventSubscription:
    """
    All event Subscription class
    """
    def __init__(self, *, token: AuthToken, filters: list):
        self.subscription_id = ID()
        self.token = token
        self.filters = {}
        for f in filters:
            self.filters[ID()] = f

    def get_subscription_id(self):
        """
        Get Subscription Id
        """
        return self.subscription_id

    def has_access(self, *, token: AuthToken) -> bool:
        """
        Returns true if the token matches
        @param token token
        """
        if token is None:
            return False
        if token.get_name() is None:
            return False
        if self.token is None:
            return False
        return self.token.get_name() == token.get_name()

    def add_event_filter(self, *, f: IEventFilter):
        """
        Add event filter
        @param f filter
        """
        fid = ID()
        self.filters[fid] = f

    def delete_filter(self, *, fid: ID):
        """
        Delete filter
        @param fid fid
        """
        if fid in self.filters:
            self.filters.pop(fid)

    def matches(self, *, event: IEvent):
        """
        Check if event matches the filters
        @param event event
        """
        if len(self.filters) == 0:
            return True
        for f in self.filters.values():
            if not f.matches(event=event):
                return False
        return True

    @abstractmethod
    def deliver_event(self, *, event: IEvent):
        """
        Deliver event
        @param event event
        """

    @abstractmethod
    def drain_events(self, *, timeout: int) -> list:
        """
        Drain events
        @param timeout timeout
        """

    @abstractmethod
    def is_abandoned(self) -> bool:
        """
        Returns true if abandoned; otherwise false
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING
from fabric.actor.core.apis.i_timer_task import ITimerTask
from fabric.actor.core.kernel.failed_rpc import FailedRPC
from fabric.actor.core.kernel.failed_rpc_event import FailedRPCEvent
from fabric.actor.core.util.rpc_exception import RPCException, RPCError

if TYPE_CHECKING:
    from fabric.actor.core.kernel.rpc_request import RPCRequest


class QueryTimeout(ITimerTask):
    """
    Query timeout
    """
    def __init__(self, *, req: RPCRequest):
        self.req = req

    def execute(self):
        """
        Process a query timeout
        """
        try:
            from fabric.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
            pending = RPCManagerSingleton.get().remove_pending_request(guid=self.req.request.get_message_id())
            if pending is not None:
                self.req.actor.get_logger().debug("Query timeout. Responding with FailedRPC RPC={}".format(self.req))
                failed = FailedRPC(e=RPCException(message="Timeout while waiting for query response",
                                                  error=RPCError.Timeout), request=self.req)
                self.req.actor.queue_event(incoming=FailedRPCEvent(actor=self.req.actor, failed=failed))
            else:
                self.req.actor.get_logger().debug("Query timeout. Query already completed RPC={}".format(self.req))
        except Exception as e:
            self.req.actor.get_logger().error("Claim timeout. RPC={} e: {}".format(self.req.reservation, e))
            self.req.actor.get_logger().error(traceback.format_exc())



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from datetime import datetime
from typing import List

from fabric.actor.core.util.resource_type import ResourceType
from fabric.message_bus.messages.reservation_mng import ReservationMng

from fabric.message_bus.messages.ticket_reservation_avro import TicketReservationAvro

from fabric.message_bus.messages.delegation_avro import DelegationAvro

from fabric.message_bus.messages.pool_info_avro import PoolInfoAvro

from fabric.actor.core.apis.i_actor import ActorType
from fabric.actor.core.common.exceptions import ManageException
from fabric.message_bus.messages.get_actors_request_avro import GetActorsRequestAvro
from fabric.message_bus.messages.proxy_avro import ProxyAvro

from fabric.actor.core.apis.i_mgmt_controller import IMgmtController
from fabric.actor.core.common.constants import Constants, ErrorCodes
from fabric.actor.core.manage.kafka.kafka_actor import KafkaActor
from fabric.actor.core.util.id import ID
from fabric.message_bus.messages.get_reservation_units_request_avro import GetReservationUnitsRequestAvro
from fabric.message_bus.messages.result_avro import ResultAvro
from fabric.message_bus.messages.unit_avro import UnitAvro


class KafkaController(KafkaActor, IMgmtController):
    def clone(self):
        return KafkaController(guid=self.management_id,
                               kafka_topic=self.kafka_topic,
                               auth=self.auth, logger=self.logger,
                               message_processor=self.message_processor,
                               producer=self.producer)

    def add_broker(self, *, broker: ProxyAvro) -> bool:
        raise ManageException(Constants.not_implemented)

    def get_brokers(self, *, broker: ID = None, id_token: str = None) -> List[ProxyAvro]:
        self.clear_last()
        status = ResultAvro()
        rret_val = None

        try:
            request = GetActorsRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.message_id = str(ID())
            request.callback_topic = self.callback_topic
            request.type = ActorType.Broker.name
            request.id_token = id_token
            request.broker_id = broker

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = message_wrapper.response.proxies
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return rret_val

    def get_pool_info(self, *, broker: ID, id_token: str) -> List[PoolInfoAvro]:
        raise ManageException(Constants.not_implemented)

    def claim_delegations(self, *, broker: ID, did: ID, id_token: str = None) -> DelegationAvro:
        raise ManageException(Constants.not_implemented)

    def reclaim_delegations(self, *, broker: ID, did: ID, id_token: str = None) -> DelegationAvro:
        raise ManageException(Constants.not_implemented)

    def get_reservation_units(self, *, rid: ID, id_token: str = None) -> List[UnitAvro]:
        self.clear_last()
        status = ResultAvro()
        rret_val = None

        try:
            request = GetReservationUnitsRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.message_id = str(ID())
            request.callback_topic = self.callback_topic
            request.reservation_id = str(rid)
            request.id_token = id_token

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = message_wrapper.response.units
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return rret_val

    def add_reservation(self, *, reservation: TicketReservationAvro) -> ID:
        raise ManageException(Constants.not_implemented)

    def add_reservations(self, *, reservations: List[ReservationMng]) ->list:
        raise ManageException(Constants.not_implemented)

    def demand_reservation(self, *, reservation: ReservationMng) -> bool:
        raise ManageException(Constants.not_implemented)

    def demand_reservation_rid(self, *, rid: ID) -> bool:
        raise ManageException(Constants.not_implemented)

    def extend_reservation(self, *, reservation: ID, new_end_time: datetime, new_units: int,
                           new_resource_type: ResourceType = None, request_properties: dict = None,
                           config_properties: dict = None) -> bool:
        raise ManageException(Constants.not_implemented)

    def extend_reservation_end_time(self, *, reservation: ID, new_end_time: datetime) -> bool:
        raise ManageException(Constants.not_implemented)

    def extend_reservation_end_time_request(self, *, reservation: ID, new_end_time: datetime,
                                            request_properties: dict) -> bool:
        raise ManageException(Constants.not_implemented)

    def extend_reservation_end_time_request_config(self, *, reservation: ID, new_end_time: datetime,
                                                   request_properties: dict, config_properties: dict) -> bool:
        raise ManageException(Constants.not_implemented)

    def modify_reservation(self, *, rid: ID, modify_properties: dict) -> bool:
        raise ManageException(Constants.not_implemented)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.util.resource_type import ResourceType


class CountsPerType:
    """
    Inner class representing counts for a given resource type.
    """
    def __init__(self, *, resource_type: ResourceType):
        self.resource_type = resource_type
        self.active = 0
        self.pending = 0
        self.expired = 0
        self.failed = 0
        self.closed = 0


class ResourceCount:
    """
    ResourceCount is a utility class used to count reservations in a ReservationSet based on the reservation
    state and resource type. An instance of this class is used as an argument to the "tally" methods in ReservationSet.
    Note: no internal locking.
    """
    def __init__(self):
        # Counts per resource type. <String, CountsPerType>
        self.map = {}

    def get_counts(self, *, resource_type: ResourceType) -> CountsPerType:
        """
        Returns the count entry for the given resource type.

        Args:
            resource_type: resource type
        Returns:
            count entry for the given resource type
        """
        return self.map.get(resource_type.get_type(), None)

    def get_or_create_counts(self, *, resource_type: ResourceType) -> CountsPerType:
        """
        Returns or creates a new count entry for the given resource type.

        Args:
            resource_type: resource type
        Returns:
            count entry for the given resource type
        """
        result = self.map.get(resource_type.get_type(), None)

        if result is None:
            result = CountsPerType(resource_type=resource_type)
            self.map[resource_type.get_type()] = result

        return result

    def count_active(self, *, resource_type: ResourceType) -> int:
        """
        Returns the number of active units from the given resource type.

        Args:
            resource_type: resource type
        Returns:
            number of active units
        """
        counts = self.get_counts(resource_type=resource_type)
        if counts is not None:
            return counts.active
        else:
            return 0

    def count_close(self, *, resource_type: ResourceType) -> int:
        """
        Returns the number of closed units from the given resource type.

        Args:
            resource_type: resource type
        Returns:
            number of closed units
        """
        counts = self.get_counts(resource_type=resource_type)
        if counts is not None:
            return counts.closed
        else:
            return 0

    def count_expired(self, *, resource_type: ResourceType) -> int:
        """
        Returns the number of expired units from the given resource type.

        Args:
            resource_type: resource type
        Returns:
            number of expired units
        """
        counts = self.get_counts(resource_type=resource_type)
        if counts is not None:
            return counts.expired
        else:
            return 0

    def count_failed(self, *, resource_type: ResourceType) -> int:
        """
        Returns the number of failed units from the given resource type.

        Args:
            resource_type: resource type
        Returns:
            number of failed units
        """
        counts = self.get_counts(resource_type=resource_type)
        if counts is not None:
            return counts.failed
        else:
            return 0

    def count_pending(self, *, resource_type: ResourceType) -> int:
        """
        Returns the number of pending units from the given resource type.

        Args:
            resource_type: resource type
        Returns:
            number of pending units
        """
        counts = self.get_counts(resource_type=resource_type)
        if counts is not None:
            return counts.pending
        else:
            return 0

    def tally_active(self, *, resource_type: ResourceType, count: int):
        """
        Increments with count the internal counter for active units of the specified type.

        Args:
            resource_type: resource type
            count: count
        """
        counts = self.get_or_create_counts(resource_type=resource_type)
        counts.active += count

    def tally_close(self, *, resource_type: ResourceType, count: int):
        """
        Increments with count the internal counter for closed units of the specified type.

        Args:
            resource_type: resource type
            count: count
        """
        counts = self.get_or_create_counts(resource_type=resource_type)
        counts.closed += count

    def tally_expired(self, *, resource_type: ResourceType, count: int):
        """
        Increments with count the internal counter for expired units of the specified type.

        Args:
            resource_type: resource type
            count: count
        """
        counts = self.get_or_create_counts(resource_type=resource_type)
        counts.expired += count

    def tally_failed(self, *, resource_type: ResourceType, count: int):
        """
        Increments with count the internal counter for failed units of the specified type.

        Args:
            resource_type: resource type
            count: count
        """
        counts = self.get_or_create_counts(resource_type=resource_type)
        counts.failed += count

    def tally_pending(self, *, resource_type: ResourceType, count: int):
        """
        Increments with count the internal counter for pending units of the specified type.

        Args:
            resource_type: resource type
            count: count
        """
        counts = self.get_or_create_counts(resource_type=resource_type)
        counts.pending += count


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import RegistryException
from fabric.actor.core.proxies.proxy import Proxy

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker_proxy import IBrokerProxy
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.apis.i_base_plugin import IBasePlugin


class PeerRegistry:
    def __init__(self):
        # A hash table of proxies to brokers. Persisted to the data store.
        self.brokers = {}
        self.default_broker = None

        # Cache of objects per actor name
        self.cache = {}
        self.plugin = None
        self.logger = None
        self.initialized = False
        self.lock = threading.Lock()

    def actor_added(self):
        self.load_from_db()

    def add_broker(self, *, broker: IBrokerProxy):
        try:
            self.lock.acquire()
            self.brokers[broker.get_identity().get_guid()] = broker

            if self.default_broker is None:
                self.default_broker = broker
        finally:
            self.lock.release()

        try:
            self.plugin.get_database().add_broker(broker=broker)
            self.logger.info("Added {} as broker".format(broker.get_name()))
        except Exception as e:
            self.plugin.get_logger().error("Error while adding broker {}".format(e))

    def get_broker(self, *, guid: ID) -> IBrokerProxy:
        ret_val = None
        try:
            self.lock.acquire()
            ret_val = self.brokers.get(guid, None)
        finally:
            self.lock.release()
        return ret_val

    def get_brokers(self) -> list:
        ret_val = []
        try:
            self.lock.acquire()
            for b in self.brokers.values():
                ret_val.append(b)
        finally:
            self.lock.release()
        return ret_val

    def get_default_broker(self) -> IBrokerProxy:
        try:
            self.lock.acquire()
            return self.default_broker
        finally:
            self.lock.release()

    def initialize(self):
        if not self.initialized:
            if self.plugin is None:
                raise RegistryException(Constants.not_specified_prefix.format("plugin"))

            self.logger = self.plugin.get_logger()

    def load_from_db(self):
        brokers = self.plugin.get_database().get_brokers()
        count = 0
        for b in brokers:
            agent = Proxy.get_proxy(properties=b)
            self.brokers[agent.get_identity().get_guid()] = agent
            if count == 0:
                self.default_broker = agent
            count += 1

    def remove_broker(self, *, broker: IBrokerProxy):
        try:
            self.lock.acquire()
            if broker.get_identity() in self.brokers:
                self.brokers.pop(broker.get_identity)
        finally:
            self.lock.release()

    def set_slices_plugin(self, *, plugin: IBasePlugin):
        self.plugin = plugin




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from datetime import datetime

from fabric.actor.core.common.exceptions import ResourcesException
from fabric.actor.core.util.resource_type import ResourceType

from fabric.actor.core.common.resource_pool_attribute_descriptor import ResourcePoolAttributeDescriptor


class ResourcePoolDescriptor:
    property_type = "type"
    property_label = "label"
    property_description = "description"
    property_attributes_prefix = "attribute."
    property_attributes_count = "attributescount"
    property_key = "key"

    def __init__(self):
        self.attributes = {}
        self.units = 0
        self.start = None
        self.end = None
        self.handler_module = None
        self.handler_class = None
        self.handler_properties = {}
        self.pool_properties = {}
        self.resource_type = None
        self.resource_label = None
        self.description = None
        self.factory_module = None
        self.factory_class = None

    def __str__(self):
        return "units:{} start:{} end:{} handler_module:{} handler_class:{} handler_properties:{} pool_properties:{} " \
               "resource_type:{} resource_label:{} description:{} factory_module:{} factory_class:{}".\
            format(self.units, self.start, self.end, self.handler_module, self.handler_class, self.handler_properties,
                   self.pool_properties, self.resource_type, self.resource_label, self.description, self.factory_module,
                   self.factory_class)

    def get_resource_type(self) -> ResourceType:
        """
        Get resource type
        @return resource type
        """
        return self.resource_type

    def set_resource_type(self, *, rtype: ResourceType):
        """
        Set resource type
        @param rtype resource type
        """
        self.resource_type = rtype

    def get_handler_module(self) -> str:
        """
        Get Handler Module
        @return handler module
        """
        return self.handler_module

    def set_handler_module(self, *, module: str):
        """
        Set Handler module
        @param module Handler module
        """
        self.handler_module = module

    def set_handler_class(self, *, handler_class: str):
        """
        Set Handler class
        @param handler_class Handler class
        """
        self.handler_class = handler_class

    def get_handler_class(self) -> str:
        """
        Get Handler class
        @return handler class
        """
        return self.handler_class

    def get_handler_properties(self) -> dict:
        """
        Get Handler Properties
        @return handler Properties
        """
        return self.handler_properties

    def set_handler_properties(self, *, properties: dict):
        """
        Set handler Properties
        @param properties handler Properties
        """
        self.handler_properties = properties

    def get_pool_properties(self) -> dict:
        """
        Get Pool Properties
        @return Pool Properties
        """
        return self.pool_properties

    def set_pool_properties(self, *, properties: dict):
        """
        Set Pool Properties
        @param properties Pool Properties
        """
        self.pool_properties = properties

    def get_resource_type_label(self) -> str:
        """
        Get resource type label
        @return resource type label
        """
        return self.resource_label

    def set_resource_type_label(self, *, rtype_label: str):
        """
        Set resource type label
        @param rtype_label resource type label
        """
        self.resource_label = rtype_label

    def get_units(self) -> int:
        """
        Get units
        @return units
        """
        return self.units

    def set_units(self, *, units: int):
        """
        Set units
        @param units units
        """
        self.units = units

    def get_start(self) -> datetime:
        """
        Get Start Lease Time
        @return Start Lease Time
        """
        return self.start

    def set_start(self, *, start: datetime):
        """
        Set start Lease Time
        @param start start Lease Time
        """
        self.start = start

    def get_end(self) -> datetime:
        """
        Get End Lease Time
        @return End Lease Time
        """
        return self.end

    def set_end(self, *, end: datetime):
        """
        Set End Lease Time
        @param end End Lease Time
        """
        self.end = end

    def get_description(self) -> str:
        """
        Get Description
        @return description
        """
        return self.description

    def set_description(self, *, description: str):
        """
        Set description
        @param description description
        """
        self.description = description

    def get_pool_factory_module(self) -> str:
        """
        Get Pool Factory Module
        @return factory module
        """
        return self.factory_module

    def set_pool_factory_module(self, *, factory_module: str):
        """
        Set Pool factory module
        @param factory_module factory module
        """
        self.factory_module = factory_module

    def get_pool_factory_class(self) -> str:
        """
        Get Pool Factory class
        @return pool factory class
        """
        return self.factory_class

    def set_pool_factory_class(self, *, factory_class: str):
        """
        Set Pool Factory Class
        @param factory_class factory class
        """
        self.factory_module = factory_class

    def get_attribute(self, *, key: str) -> ResourcePoolAttributeDescriptor:
        """
        Get an attribute
        @param key attribute key
        """
        if key in self.attributes:
            return self.attributes[key]
        return None

    def get_attributes(self) -> list:
        """
        Return attributes
        @return list of attributes
        """
        return [x for x in self.attributes.values()]

    def add_attribute(self, *, attribute: ResourcePoolAttributeDescriptor):
        """
        Add an attribute
        @param attribute Attribute
        """
        self.attributes[attribute.get_key()] = attribute

    def save(self, *, properties: dict, prefix: str = None) -> dict:
        """
        Save properties
        @param properties properties
        @param prefix prefix
        @return properties
        """
        if prefix is None:
            prefix = ""

        properties[prefix + self.property_type] = str(self.resource_type)
        properties[prefix + self.property_label] = self.resource_label
        if self.description is not None:
            properties[prefix + self.property_description] = self.description

        properties[prefix + self.property_attributes_count] = str(len(self.attributes))
        i = 0
        for a in self.attributes.values():
            properties[prefix + self.property_attributes_prefix + str(i) + "." + self.property_key] = a.get_key()
            if len(prefix) > 0:
                temp = prefix + a.get_key() + "."
            else:
                temp = a.get_key() + "."
            properties = a.save(properties=properties, prefix=temp)
            i += 1
        return properties

    def reset(self, *, properties: dict, prefix: str = None):
        """
        Reset properties
        @param properties properties
        @param prefix prefix
        """
        if prefix is None:
            prefix = ""

        if (prefix + self.property_type) not in properties:
            raise ResourcesException("Missing resource type")

        self.resource_type = ResourceType(resource_type=properties[prefix + self.property_type])

        if (prefix + self.property_label) not in properties:
            raise ResourcesException("Missing resource label")

        self.resource_label = properties[prefix + self.property_label]

        if prefix + self.property_description in properties:
            self.description = properties[prefix + self.property_description]

        if (prefix + self.property_attributes_count) not in properties:
            raise ResourcesException("Missing attributes count")

        count = int(properties[prefix + self.property_attributes_count])
        for i in range(count):
            key = prefix + self.property_attributes_prefix + str(i) + "." + self.property_key
            if key not in properties:
                raise ResourcesException("Could not find key for attribute #{}".format(i))
            key_value = properties[key]

            if len(prefix) > 0:
                temp = prefix + key_value + "."
            else:
                temp = key_value + "."

            attribute = ResourcePoolAttributeDescriptor()
            attribute.reset(properties=properties, prefix=temp)
            attribute.set_key(value=key_value)
            self.add_attribute(attribute=attribute)

    def clone(self):
        """
        Clone an object
        @return copy of current object
        """
        properties = {}
        self.save(properties=properties)
        copy = ResourcePoolDescriptor()
        try:
            copy.reset(properties=properties)
        except Exception as e:
            raise ResourcesException("Unexpected error during deserialization={}".format(e))
        return copy



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING
from fabric.actor.core.policy.free_allocated_set import FreeAllocatedSet
from fabric.actor.core.policy.inventory_for_type import InventoryForType

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_client_reservation import IClientReservation


class SimplerUnitsInventory(InventoryForType):
    def __init__(self):
        super().__init__()
        self.set = FreeAllocatedSet()

    def donate(self, *, source: IClientReservation):
        super().donate(source=source)
        rset = source.get_resources()
        cset = rset.get_resources()
        ticket = cset.get_ticket()

        for i in range(ticket.get_units()):
            self.set.add_inventory(item=i)

    def allocate(self, *, count: int, request: dict, resource: dict = None) -> dict:
        self.set.allocate(count=count)
        result = {}
        return result

    def allocate_revisit(self, *, count: int, resource: dict):
        self.set.allocate(count=count)

    def free(self, *, count: int, request: dict = None, resource: dict = None) -> dict:
        self.set.free(count=count)
        result = {}
        return result

    def get_free(self) -> int:
        return self.set.get_free()

    def get_allocated(self) -> int:
        return self.set.get_allocated()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time
import unittest

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_authority import IAuthority
from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.apis.i_slice import ISlice
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.policy.authority_calendar_policy import AuthorityCalendarPolicy
from fabric.actor.core.policy.resource_control import ResourceControl
from fabric.actor.core.policy.vlan_control import VlanControl
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.util.update_data import UpdateData
from fabric.actor.test.core.policy.authority_calendar_policy_test import AuthorityCalendarPolicyTest


class VlanControlTest(AuthorityCalendarPolicyTest, unittest.TestCase):
    StartVlan = 50
    EndVlan = StartVlan + AuthorityCalendarPolicyTest.DonateUnits - 1
    VlanBW = 100

    from fabric.actor.core.container.globals import Globals
    Globals.config_file = Constants.test_vm_am_configuration_file

    from fabric.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def get_control(self, policy: AuthorityCalendarPolicy = None):
        if policy:
            return policy.controls_by_guid.values().__iter__().__next__()

        control = VlanControl()
        properties = {ResourceControl.PropertyControlResourceTypes: str(AuthorityCalendarPolicyTest.Type)}
        control.configure(properties=properties)
        return control

    def get_source(self, units: int, rtype: ResourceType, term: Term, actor: IActor, slice_obj: ISlice):
        resources = ResourceSet(units=units, rtype=rtype, rdata=ResourceData())

        local = resources.get_local_properties()
        local[Constants.property_vlan_range_num] = "1"
        local[Constants.property_start_vlan + "0"] = str(self.StartVlan)
        local[Constants.property_end_vlan + "0"] = str(self.EndVlan)

        properties = {Constants.resource_bandwidth: str(self.VlanBW)}

        delegation = actor.get_plugin().get_ticket_factory().make_delegation(units=units, term=term, rtype=rtype,
                                                                             properties=properties, vector=None)
        ticket = actor.get_plugin().get_ticket_factory().make_ticket(delegation=delegation, source=None)
        cs = Ticket(ticket=ticket, plugin=actor.get_plugin())
        resources.set_resources(cset=cs)
        source = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        return source

    def check_before_donate(self, authority: IAuthority):
        super().check_before_donate(authority)
        policy = authority.get_policy()
        control = self.get_control(policy)
        self.assertIsNotNone(control.tags)
        self.assertEqual(0, control.tags.size())

    def check_after_donate(self, authority: IAuthority, source: IClientReservation):
        super().check_after_donate(authority, source)
        policy = authority.get_policy()
        control = self.get_control(policy)

        self.assertEqual(self.DonateUnits, control.tags.size())
        self.assertEqual(self.DonateUnits, control.tags.get_free())
        self.assertEqual(0, control.tags.get_allocated())
        self.assertEqual(self.Type, control.rtype)

    vlan_tag = 0

    def check_incoming_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation,
                             udd: UpdateData):
        super().check_incoming_lease(authority, request, incoming, udd)
        policy = authority.get_policy()
        control = self.get_control(policy)

        rset = incoming.get_resources()
        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(self.TicketUnits, uset.get_units())

        u = uset.get_set().values().__iter__().__next__()
        stag = u.get_property(name=Constants.unit_vlan_tag)
        self.assertIsNotNone(stag)
        self.vlan_tag = int(stag)
        self.assertEqual(self.DonateUnits - 1, control.tags.get_free())
        self.assertEqual(1, control.tags.get_allocated())

    def check_incoming_close_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation,
                                   udd: UpdateData):
        super().check_incoming_close_lease(authority, request, incoming, udd)
        policy = authority.get_policy()
        control = self.get_control(policy)

        rset = incoming.get_resources()
        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(0, uset.get_units())
        self.assertEqual(self.DonateUnits, control.tags.get_free())
        self.assertEqual(0, control.tags.get_allocated())

    def check_incoming_extend_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation,
                                    update_data: UpdateData):
        super().check_incoming_extend_lease(authority, request, incoming, update_data)
        policy = authority.get_policy()
        control = self.get_control(policy)

        rset = incoming.get_resources()
        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(self.TicketUnits, uset.get_units())

        u = uset.get_set().values().__iter__().__next__()
        stag = u.get_property(Constants.unit_vlan_tag)
        self.assertIsNotNone(stag)
        self.assertEqual(self.vlan_tag, int(stag))
        self.assertEqual(self.DonateUnits - 1, control.tags.get_free())
        self.assertEqual(1, control.tags.get_allocated())

        self.assertEqual(self.DonateUnits, control.tags.get_free())
        self.assertEqual(0, control.tags.get_allocated())




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.kernel.incoming_rpc import IncomingRPC

if TYPE_CHECKING:
    from fabric.actor.core.kernel.rpc_request_type import RPCRequestType
    from fabric.actor.core.util.id import ID
    from fabric.actor.security.auth_token import AuthToken


class IncomingFailedRPC(IncomingRPC):
    """
    Represents an incoming failed rpc message
    """
    def __init__(self, message_id: ID, failed_request_type: RPCRequestType, request_id: str,
                 failed_reservation_id: ID, error_details: str, caller: AuthToken):
        super().__init__(message_id=message_id, request_type=RPCRequestType.FailedRPC, caller=caller,
                         callback=None)
        self.failed_request_type = failed_request_type
        self.failed_reservation_id = failed_reservation_id
        self.error_details = error_details
        self.request_id = request_id

    def get_error_details(self) -> str:
        """
        Get Error details
        @return error details
        """
        return self.error_details

    def get_failed_request_type(self) -> RPCRequestType:
        """
        Get failed request type
        @return failed request type
        """
        return self.failed_request_type

    def get_failed_reservation_id(self) -> ID():
        """
        Get failed reservation id
        @return failed reservation id
        """
        return self.failed_reservation_id



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List
from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.kernel.reservation_states import JoinState

if TYPE_CHECKING:
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.kernel.resource_set import ResourceSet


class IControllerReservation(IClientReservation):
    """
    Interface for Controller/Orchestrator Reservations
    """
    @abstractmethod
    def get_join_state(self) -> JoinState:
        """
        Returns the join state.
        @returns join state
        """

    @abstractmethod
    def get_join_state_name(self) -> str:
        """
        Returns the name of the join state.
        @returns name of the join state
        """

    @abstractmethod
    def get_leased_resources(self) -> ResourceSet:
        """
        Returns the resources leased by the reservation. If the reservation has
        not yet issued a redeem request, returns null.
        @returns resources leased by the reservation. Can be null.
        """

    @abstractmethod
    def get_lease_sequence_in(self) -> int:
        """
        Returns the reservation sequence number for incoming ticket/extend ticket messages.
        @returns reservation sequence number for incoming ticket/extend ticket messages
        """

    @abstractmethod
    def get_lease_sequence_out(self) -> int:
        """
        Returns the reservation sequence number for outgoing ticket/extend ticket messages.
        @returns reservation sequence number for outgoing ticket/extend ticket messages
        """

    @abstractmethod
    def get_lease_term(self) -> Term:
        """
        Returns the term of the current lease.
        @returns current lease term
        """

    @abstractmethod
    def get_previous_lease_term(self) -> Term:
        """
        Returns the previous lease term.
        @returns previous lease term
        """

    @abstractmethod
    def is_active_joined(self) -> bool:
        """
        Returns true if this reservation is currently active, and has completed
        joining the guest, i.e., successor reservations (with join dependencies
        on this reservation) may now join. Note: if this reservation is closed or
        failed, activeJoined returns false and successors will remain blocked,
        i.e, the caller must close them.
        @returns true or false as explained above
        """

    @abstractmethod
    def set_exported(self, *, exported: bool):
        """
        Indicates whether the reservation represents exported resources.
        @params exported value for the exported flag
        """

    @abstractmethod
    def add_join_predecessor(self, *, predecessor, filter: dict = None):
        """
        Sets the join predecessor: the reservation, for which the kernel must
        issue a join before a join may be issued for the current reservation.
        @params predecessor predecessor reservation
        @params filter: filter
        """

    @abstractmethod
    def set_lease_sequence_in(self, *, sequence: int):
        """
        Sets the reservation sequence number for incoming lease messages.
        @params sequence sequence number
        """

    @abstractmethod
    def set_lease_sequence_out(self, *, sequence: int):
        """
        Sets the reservation sequence number for outgoing lease messages.
        @params sequence sequence number
        """

    @abstractmethod
    def add_redeem_predecessor(self, *, reservation, filter: dict = None):
        """
        Adds a redeem predecessor to this reservation: the passed in reservation
        must be redeemed before this reservation.
        @params reservation: predecessor reservation
        @params filter: filter
        """

    @abstractmethod
    def get_redeem_predecessors(self) -> List[IControllerReservation]:
        """
        Returns the redeem predecessors list for the reservation.
        @returns redeem predecessors list for the reservation
        """

    @abstractmethod
    def get_join_predecessors(self) -> List[IControllerReservation]:
        """
        Returns the join predecessors list for the reservation.
        @returns join predecessors list for the reservation
        """

    @abstractmethod
    def set_config_property(self, *, key: str, value: str):
        """
        Sets a configuration property.
        @params key: key
        @params value: value
        """

    @abstractmethod
    def set_request_property(self, *, key: str, value: str):
        """
        Sets a request property.
        @params key: key
        @params value: value
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)


class DelegationNotFoundException(Exception):
    """
    Delegation Not Found Exception
    """
    def __init__(self, *, text: str = None, did=None):
        super(DelegationNotFoundException, self).__init__()
        if text is not None:
            self.text = str(text)
        else:
            self.text = "Delegation# {} not found".format(did)
        self.did = did

    def __str__(self):
        return self.text


class ReservationNotFoundException(Exception):
    """
    Reservation not found exception
    """
    def __init__(self, *, text: str = None, rid=None):
        super(ReservationNotFoundException, self).__init__()
        if text is not None:
            self.text = str(text)
        else:
            self.text = "Reservation# {} not found".format(rid)
        self.rid = rid

    def __str__(self):
        return self.text


class SliceNotFoundException(Exception):
    """
    Slice not found exception
    """
    def __init__(self, *, text: str = None, slice_id=None):
        super(SliceNotFoundException, self).__init__()
        if text is not None:
            self.text = str(text)
        else:
            self.text = "Slice# {} not found".format(slice_id)
        self.slice_id = slice_id

    def __str__(self):
        return self.text


class AuthorityException(Exception):
    """
    Authority Exception
    """


class KernelException(Exception):
    """
    Kernel Exception
    """


class ResourcesException(Exception):
    """
    Resources Exception
    """


class BrokerException(Exception):
    """
    Authority Exception
    """


class DatabaseException(Exception):
    """
    Database exception
    """


class EventException(Exception):
    """
    Event exception
    """


class ContainerException(Exception):
    """
    Exception raised by container
    """


class InitializationException(Exception):
    """
    Exception raised for Initialization Errors
    """


class KafkaServiceException(Exception):
    """
    Exception raised for Kafka Service Errors
    """


class ActorException(Exception):
    """
    Actor Exception
    """


class ControllerException(Exception):
    """
    Controller Exception
    """


class TicketException(Exception):
    """
    Ticket Exception
    """


class UnitException(Exception):
    """
    Unit Exception
    """


class DelegationException(Exception):
    """
    Delegation Exception
    """


class PluginException(Exception):
    """
    Plugin Exception
    """


class ReservationException(Exception):
    """
    Reservation Exception
    """


class SliceException(Exception):
    """
    Slice Exception
    """


class TickException(Exception):
    """
    Tick Exception
    """


class ManageException(Exception):
    """
    Manage Exception
    """


class PolicyException(Exception):
    """
    Policy Exception
    """


class ProxyException(Exception):
    """
    Proxy Exception
    """


class RegistryException(Exception):
    """
    Registry Exception
    """


class TimeException(Exception):
    """
    Time Exception
    """


class FrameworkException(Exception):
    """
    Framework Exception
    """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest

import time
from fabric.actor.core.apis.i_database import IDatabase
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.util.id import ID
from fabric.actor.test.base_test_case import BaseTestCase


class ActorDatabaseTest(BaseTestCase, unittest.TestCase):
    from fabric.actor.core.container.globals import Globals
    Globals.config_file = Constants.test_vm_am_configuration_file

    from fabric.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def get_clean_database(self) -> IDatabase:
        db = self.get_actor_database()
        db.set_actor_name(name=self.actor_name)
        db.set_reset_state(state=True)
        db.initialize()
        return db

    def test_a_create(self):
        self.get_clean_database()

    def prepare_actor_database(self):
        container_db = self.get_container_database()
        actor = self.get_actor()
        container_db.remove_actor(actor_name=actor.get_name())
        container_db.add_actor(actor=actor)
        actor.actor_added()
        return actor

    def test_b_create_2(self):
        self.prepare_actor_database()

    def get_database_to_test(self):
        actor = self.prepare_actor_database()
        return actor.get_plugin().get_database()

    def test_c_add_slice(self):
        db = self.get_database_to_test()
        slice_obj = SliceFactory.create(slice_id=ID(), name="slice_to_add")
        self.assertEqual("slice_to_add", slice_obj.get_name())
        db.add_slice(slice_object=slice_obj)
        result = db.get_slice(slice_id=slice_obj.get_slice_id())
        self.assertIsNotNone(result)
        slice2 = SliceFactory.create_instance(properties=result)
        self.assertIsNotNone(slice2)
        db.remove_slice(slice_id=slice_obj.get_slice_id())


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from enum import Enum
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy
from fabric.actor.core.apis.i_policy import IPolicy
from fabric.actor.core.util.update_data import UpdateData
from fim.graph.abc_property_graph import ABCPropertyGraph

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.apis.i_slice import ISlice


class DelegationState(Enum):
    """
    Enumeration for Delegation State
    """
    Nascent = 1
    Delegated = 2
    Closed = 3
    Reclaimed = 4
    Failed = 5


class IDelegation(ABC):
    """
    IDelegation defines the the core API for a delegation. Most of the methods described in the interface allow the
    programmer to inspect the state of the delegation, access some of its core objects,
    and wait for the occurrence of a particular event.
    """

    @abstractmethod
    def set_graph(self, graph: ABCPropertyGraph):
        """
        Set the Graph associated
        @param graph: graph
        """

    @abstractmethod
    def get_graph(self) -> ABCPropertyGraph:
        """
        Return the Graph associated
        """

    @abstractmethod
    def get_actor(self) -> IActor:
        """
        Returns the actor in control of the delegation.

        Returns:
            the actor in control of the delegation.
        """

    @abstractmethod
    def get_delegation_id(self) -> ID:
        """
        Returns the delegation id.

        Returns:
            the delegation id.
        """

    @abstractmethod
    def get_slice_object(self) -> ISlice:
        """
        Returns the slice the delegation belongs to.

        Returns:
            slice the delegation belongs to.
        """

    @abstractmethod
    def get_slice_id(self) -> ID:
        """
        Returns the resource model GUID.

        Returns:
            resource model  guid
        """

    @abstractmethod
    def get_state(self) -> DelegationState:
        """
        Returns the current delegation state.

        Returns:
            current delegation state.
        """

    @abstractmethod
    def get_state_name(self) -> str:
        """
        Returns the current delegation state name.

        Returns:
            current delegation state name.
        """

    @abstractmethod
    def set_slice_object(self, *, slice_object: ISlice):
        """
        Sets the slice the delegation belongs to.

        Args:
            slice_object: slice_object the delegation belongs to
        """

    @abstractmethod
    def set_logger(self, *, logger):
        """
        Sets the logger
        @param logger: logger
        """

    @abstractmethod
    def transition(self, *, prefix: str, state: DelegationState):
        """
        Transitions this delegation into a new state.

        Args:
            prefix: prefix
            state: the new state
        """
        raise NotImplementedError("Should have implemented this")

    @abstractmethod
    def get_notices(self) -> str:
        """
        Returns the error message associated with this delegation.
        @return error message associated with this delegation
        """

    @abstractmethod
    def is_dirty(self) -> bool:
        """
        Checks if the delegation has uncommitted updates.

        Returns:
            true if the delegation has an uncommitted updates
        """

    @abstractmethod
    def set_dirty(self):
        """
        Marks the delegation as containing uncommitted updates.
        """

    @abstractmethod
    def clear_dirty(self):
        """
        Marks that the delegation has no uncommitted updates or state transitions.
        """

    @abstractmethod
    def has_uncommitted_transition(self) -> bool:
        """
        Checks if the delegation has uncommitted state transitions.

        Returns:
            true if the delegation has an uncommitted transition
        """

    @abstractmethod
    def set_actor(self, actor: IActor):
        """
        Set Actor
        @param actor: actor
        """

    @abstractmethod
    def prepare(self, *, callback: ICallbackProxy, logger):
        """
        Prepare the delegation
        @param callback callback
        @param logger logger
        """

    @abstractmethod
    def is_closed(self) -> bool:
        """
        Returns true if delegation is closed; false otherwise
        """

    @abstractmethod
    def delegate(self, policy: IPolicy, id_token: str = None):
        """
        Check if delegation can be delegated and state transition
        @param policy policy
        @param id_token id token
        """

    @abstractmethod
    def claim(self):
        """
        Claim the delegation
        """

    @abstractmethod
    def reclaim(self, id_token: str = None):
        """
        Reclaim the delegation
        @param id_token id token
        """

    @abstractmethod
    def close(self):
        """
        Close the delegation
        """

    @abstractmethod
    def probe_pending(self):
        """
        Probe a delegation with a pending request. On server, if the
        operation completed, handle it and generate an update. If no pending
        request completed then do nothing.

        @throws Exception in case of error
        """

    @abstractmethod
    def prepare_probe(self):
        """
        Prepares a delegation probe.

        @throws Exception in case of error
        """

    @abstractmethod
    def service_probe(self):
        """
        Finishes processing probe.
        @throws Exception in case of error
        """

    @abstractmethod
    def get_callback(self) -> ICallbackProxy:
        """
        Returns the callback proxy.
        @return callback proxy
        """

    @abstractmethod
    def get_update_data(self) -> UpdateData:
        """
        Returns data to be sent back to the client in an update message.
        @return data to be sent back to the client in an update message
        """

    @abstractmethod
    def get_sequence_in(self):
        """
        Returns the sequence number of the last received message.
        @returns sequence number of the last received message
        """

    @abstractmethod
    def get_sequence_out(self):
        """
        Returns the sequence number of the last sent message.
        @returns sequence number of the last sent message
        """

    @abstractmethod
    def restore(self, actor: IActor, slice_obj: ISlice, logger):
        """
        Restore a reservation after reading from database
        @param actor: actor
        @param slice_obj: slice object
        @param logger: logger
        """

    @abstractmethod
    def validate_incoming(self):
        """
        Validate an incoming delegation
        @raises Exception in case of failure
        """

    @abstractmethod
    def service_delegate(self):
        """
        Finishes processing delegation.
        @throws Exception in case of error
        """

    @abstractmethod
    def validate_outgoing(self):
        """
        Validate an outgoing delegation
        @raises Exception in case of failure
        """

    @abstractmethod
    def service_update_delegation(self):
        """
        Finishes processing update delegation.
        @throws Exception in case of error
        """

    @abstractmethod
    def update_delegation(self, *, incoming: IDelegation, update_data: UpdateData):
        """
        Handles an incoming delegation update.

        @param incoming incoming delegation update
        @param update_data update data

        @throws Exception in case of error
        """

    @abstractmethod
    def load_graph(self, *, graph_str: str):
        """
        Load Neo4j graph from string
        @param graph_str: graph_str
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_broker_proxy import IBrokerProxy
from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_client_callback_proxy import IClientCallbackProxy
from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_rpc_request_state import IRPCRequestState
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.delegation.broker_delegation_factory import BrokerDelegationFactory
from fabric.actor.core.kernel.broker_reservation_factory import BrokerReservationFactory
from fabric.actor.core.proxies.local.local_proxy import LocalProxy
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.security.auth_token import AuthToken


class LocalBroker(LocalProxy, IBrokerProxy):
    """
    Local proxy for Broker. Allows communication with a Broker in the same container as the caller.
    """
    def __init__(self, *, actor: IActor):
        super().__init__(actor=actor)

    def _prepare_delegation(self, *, delegation: IDelegation, callback: IClientCallbackProxy,
                            caller: AuthToken, id_token: str = None) -> IRPCRequestState:
        state = LocalProxy.LocalProxyRequestState()
        state.delegation = self.pass_delegation(delegation=delegation, auth=caller)
        state.callback = callback
        return state

    def prepare_claim_delegation(self, *, delegation: IDelegation, callback: IClientCallbackProxy,
                                 caller: AuthToken, id_token: str = None) -> IRPCRequestState:
        return self._prepare_delegation(delegation=delegation, callback=callback, caller=caller, id_token=id_token)

    def prepare_reclaim_delegation(self, *, delegation: IDelegation, callback: IClientCallbackProxy,
                                   caller: AuthToken, id_token: str = None) -> IRPCRequestState:
        return self._prepare_delegation(delegation=delegation, callback=callback, caller=caller, id_token=id_token)

    def _prepare(self, *, reservation: IReservation, callback: IClientCallbackProxy,
                 caller: AuthToken) -> IRPCRequestState:
        state = LocalProxy.LocalProxyRequestState()
        state.reservation = self.pass_reservation(reservation=reservation, auth=caller)
        state.callback = callback
        return state

    def prepare_ticket(self, *, reservation: IReservation, callback: IClientCallbackProxy,
                       caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_extend_ticket(self, *, reservation: IReservation, callback: IClientCallbackProxy,
                              caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_relinquish(self, *, reservation: IReservation, callback: IClientCallbackProxy,
                           caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def pass_reservation(self, *, reservation: IClientReservation, auth: AuthToken) -> IBrokerReservation:
        slice_obj = reservation.get_slice().clone_request()

        rset = self.abstract_clone_broker(rset=reservation.get_requested_resources())
        rset.get_resource_data().request_properties = ResourceData.merge_properties(
            from_props=reservation.get_slice().get_request_properties(),
            to_props=rset.get_resource_data().get_request_properties())

        term = reservation.get_requested_term().clone()

        broker_reservation = BrokerReservationFactory.create(rid=reservation.get_reservation_id(), resources=rset,
                                                             term=term, slice_obj=slice_obj)
        broker_reservation.set_sequence_in(sequence=reservation.get_ticket_sequence_out())
        broker_reservation.set_owner(owner=self.get_identity())

        return broker_reservation

    def pass_delegation(self, *, delegation: IDelegation, auth: AuthToken) -> IDelegation:
        slice_obj = delegation.get_slice_object().clone_request()

        broker_delegation = BrokerDelegationFactory.create(did=str(delegation.get_delegation_id()),
                                                           slice_id=slice_obj.get_slice_id(),
                                                           broker=self)
        broker_delegation.set_sequence_in(sequence=delegation.get_sequence_out())
        broker_delegation.set_owner(owner=self.get_identity())

        return broker_delegation



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import pickle
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import DatabaseException
from fabric.actor.core.extensions.plugin import Plugin
from fabric.actor.core.apis.i_actor import IActor, ActorType
from fabric.actor.core.apis.i_container_database import IContainerDatabase
from fabric.actor.db.psql_database import PsqlDatabase

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_management_object import IManagementObject
    from fabric.actor.core.util.id import ID


class ContainerDatabase(IContainerDatabase):
    """
    Implements Container Interface to various Database operations
    """
    PropertyTime = "time"
    PropertyContainer = "container"

    def __init__(self, *, user: str, password: str, database: str, db_host: str, logger):
        self.user = user
        self.password = password
        self.database = database
        self.db_host = db_host
        self.db = PsqlDatabase(user=user, password=password, database=database, db_host=db_host, logger=logger)
        self.initialized = False
        self.reset_state = False
        self.logger = logger

    def __getstate__(self):
        state = self.__dict__.copy()
        self.db = PsqlDatabase(user=self.user, password=self.password, database=self.database, db_host=self.db_host, logger=self.logger)
        del state['initialized']
        del state['reset_state']
        del state['logger']

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.initialized = False
        self.reset_state = False

    def initialize(self):
        """
        Initialize
        """
        if not self.initialized:
            if self.reset_state:
                self.db.create_db()
                self.db.reset_db()
            self.initialized = True

    def set_reset_state(self, *, value: bool):
        """
        Set Reset State
        """
        self.reset_state = value

    def reset_db(self):
        """
        Reset the database
        """
        self.db.reset_db()

    def add_actor(self, *, actor: IActor):
        """
        Add an actor
        @param actor actor
        """
        properties = pickle.dumps(actor)
        self.db.add_actor(name=actor.get_name(), guid=str(actor.get_guid()), act_type=actor.get_type().value,
                          properties=properties)

    def remove_actor(self, *, actor_name: str):
        """
        Remove an actor
        @param actor_name actor name
        """
        self.db.remove_actor(name=actor_name)

    def remove_actor_database(self, *, actor_name: str):
        """
        Remove an actor
        @param actor_name actor name
        """
        self.db.remove_actor(name=actor_name)

    def get_actors(self, *, name: str = None, actor_type: int = None) -> list:
        """
        Get Actors
        @param name actor name
        @param actor_type actor type
        @return list of actors
        """
        result = None
        try:
            if name is None and actor_type is None:
                result = self.db.get_actors()
            elif name is not None and actor_type is not None:
                name = "%{}%".format(name)
                if actor_type != ActorType.All.value:
                    result = self.db.get_actors_by_name_and_type(actor_name=name, act_type=actor_type)
                else:
                    result = self.db.get_actors_by_name(act_name=name)
        except Exception as e:
            self.logger(e)
        return result

    def get_actor(self, *, actor_name: str) -> dict:
        """
        Get Actor
        @param name actor name
        @return actor
        """
        result = None
        try:
            result = self.db.get_actor(name=actor_name)
        except Exception as e:
            self.logger(e)
        return result

    def add_time(self, *, properties: dict):
        """
        Add time
        @param properties properties
        """
        self.db.add_miscellaneous(name=self.PropertyTime, properties=properties)

    def get_time(self) -> dict:
        """
        Get Time
        @param time properties
        """
        result = None
        try:
            result = self.db.get_miscellaneous(name=self.PropertyTime)
        except Exception as e:
            self.logger(e)
        return result

    def add_container_properties(self, *, properties: dict):
        """
        Add container properties
        @param properties properties
        """
        self.db.add_miscellaneous(name=self.PropertyContainer, properties=properties)

    def get_container_properties(self) -> dict:
        """
        Get Container Properties
        @return properties
        """
        result = None
        try:
            result = self.db.get_miscellaneous(name=self.PropertyContainer)
        except Exception as e:
            self.logger(e)
        return result

    def _get_plugin_from_dict(self, *, plug_obj: dict) -> Plugin:
        """
        Get Plugin from database row
        @param plug_obj plugin object read from database
        """
        if Constants.property_pickle_properties not in plug_obj:
            raise DatabaseException(Constants.invalid_argument)

        serialized_plugin = plug_obj[Constants.property_pickle_properties]
        deserialized_plugin = pickle.loads(serialized_plugin)
        return deserialized_plugin

    def get_plugin(self, *, plugin_id: str) -> Plugin:
        """
        Returns the specified plugin.
        @param plugin_id plugin identifier
        @return returns the specified plugin
        """
        result = None
        try:
            plug_obj = self.db.get_plugin(plugin_id=plugin_id)
            result = self._get_plugin_from_dict(plug_obj=plug_obj)
        except Exception as e:
            self.logger.error(e)
        return result

    def get_plugins(self, *, plugin_type: int, actor_type: int) -> list:
        """
        Returns an array of installed plugins.
        @param plugin_type plugin type
        @param actor_type plugin actor type
        @return returns the list of installed plugins
        """
        result = []
        try:
            for p in self.db.get_plugins(plg_type=plugin_type, plg_actor_type=actor_type):
                plugin = self._get_plugin_from_dict(plug_obj=p)
                result.append(plugin)
        except Exception as e:
            self.logger.error(e)
        return result

    def add_plugin(self, *, plugin: Plugin):
        """
        Add plugin
        @param plugin plugin
        """
        properties = pickle.dumps(plugin)
        self.db.add_plugin(plugin_id=plugin.get_id(), plg_type=plugin.get_plugin_type(),
                           plg_actor_type=plugin.get_actor_type(), properties=properties)

    def remove_plugin(self, *, plugin_id: str):
        """
        Remove plugin
        @param plugin_id plugin id
        """
        self.db.remove_plugin(plugin_id=plugin_id)

    def get_manager_objects_by_actor_name(self, *, actor_name: str) -> list:
        """
        Get Management Object by actor name
        @param actor_name actor name
        @return list of management objects
        """
        result = None
        try:
            result = self.db.get_manager_objects_by_actor_name(act_name=actor_name)
        except Exception as e:
            self.logger(e)
        return result

    def get_manager_containers(self) -> list:
        """
        Get Management Container
        @return list of management objects for containers
        """
        result = None
        try:
            result = self.db.get_manager_containers()
        except Exception as e:
            self.logger(e)
        return result

    def add_manager_object(self, *, manager: IManagementObject):
        """
        Add Management object
        @param manager management object
        """
        properties = manager.save()
        self.db.add_manager_object(manager_key=str(manager.get_id()), properties=properties)

    def remove_manager_object(self, *, mid: ID):
        """
        Remove management object
        @param mid management object id
        """
        self.db.remove_manager_object(manager_key=str(mid))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.core.unit_set import UnitSet
from fabric.actor.core.core.pool_manager import PoolManager
from fabric.actor.core.plugins.substrate.substrate import Substrate

if TYPE_CHECKING:
    from fabric.actor.core.core.actor import Actor
    from fabric.actor.core.plugins.config.config import Config
    from fabric.actor.core.apis.i_substrate_database import ISubstrateDatabase
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.apis.i_slice import ISlice


class AuthoritySubstrate(Substrate):
    def __init__(self, *, actor: Actor, db: ISubstrateDatabase, config: Config):
        super().__init__(actor=actor, db=db, config=config)
        self.pool_manager = None
        self.initialized = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['ticket_factory']
        del state['actor']
        del state['initialized']
        del state['pool_manager']

        del state['db']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.ticket_factory = None
        self.actor = None
        self.initialized = None
        self.pool_manager = None

    def initialize(self):
        if not self.initialized:
            super().initialize()
            self.pool_manager = PoolManager(db=self.get_database(), identity=self.actor,
                                            logger=self.get_logger())
            self.initialized = True

    def get_pool_manager(self) -> PoolManager:
        return self.pool_manager

    def revisit(self, *, slice_obj: ISlice = None, reservation: IReservation = None,
                delegation: IDelegation = None):
        if slice_obj is not None and slice_obj.is_inventory():
            self.recover_inventory_slice(slice_obj=slice_obj)

    def recover_inventory_slice(self, *, slice_obj: ISlice):
        return

    def get_units(self, *, slice_obj: ISlice) -> UnitSet:
        # TODO recovery from database
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_actor import ActorType
from fabric.actor.core.common.constants import Constants
from fabric.actor.security.fabric_token import FabricToken
from fabric.actor.security.pdp_auth import ActionId, ResourceType, PdpAuth


class AccessChecker:
    """
    Check access for Incoming operation against Policy Decision Point PDP
    """
    @staticmethod
    def check_access(*, action_id: ActionId, resource_type: ResourceType, token: str,
                     resource_id: str = None, logger=None, actor_type: ActorType) -> bool:
        """
        Check access for Incoming operation against Policy Decision Point PDP
        @param action_id action id
        @param resource_type resource type
        @param token fabric token
        @param resource_id resource id
        @param logger logger
        @param actor_type actor type
        """
        from fabric.actor.core.container.globals import GlobalsSingleton
        token_public_key = GlobalsSingleton.get().get_config().get_oauth_config().get(
            Constants.property_conf_o_auth_token_public_key, None)
        pdp_config = GlobalsSingleton.get().get_config().get_global_config().get_pdp_config()

        fabric_token = FabricToken(token_public_key=token_public_key, logger=logger,
                                   token=token)
        fabric_token.validate()
        pdp_auth = PdpAuth(config=pdp_config, logger=logger)
        return pdp_auth.check_access(fabric_token=fabric_token.get_decoded_token(),
                                     actor_type=actor_type,
                                     action_id=action_id, resource_type=resource_type,
                                     resource_id=resource_id)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List

from fabric.actor.core.apis.i_component import IComponent


if TYPE_CHECKING:
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
    from fabric.actor.core.apis.i_mgmt_controller import IMgmtController
    from fabric.actor.core.apis.i_mgmt_broker import IMgmtBroker
    from fabric.actor.core.apis.i_mgmt_authority import IMgmtAuthority
    from fabric.message_bus.messages.actor_avro import ActorAvro
    from fabric.message_bus.messages.proxy_avro import ProxyAvro


class IMgmtContainer(IComponent):
    """
    Interface for Management Container
    """
    @abstractmethod
    def get_actor(self, *, guid: ID) -> IMgmtActor:
        """
        Obtains the specified actor.
        @param guid
                   actor guid
        @return proxy to the actor on success, null otherwise. The returned proxy
        """

    @abstractmethod
    def get_controller(self, *, guid: ID) -> IMgmtController:
        """
        Obtains the specified orchestrator.
        @param guid guid
        @return specified orchestrator
        """

    @abstractmethod
    def get_broker(self, *, guid: ID) -> IMgmtBroker:
        """
        Obtains the specified broker
        @param guid guid
        @return specified broker
        """

    @abstractmethod
    def get_authority(self, *, guid: ID) -> IMgmtAuthority:
        """
        Obtains the specified authority
        @param guid guid
        @return specified authority
        """

    @abstractmethod
    def get_actors(self) -> List[ActorAvro]:
        """
        Obtains a list of all active actors in the container.
        @return list of actors. Always non-null.
        """

    @abstractmethod
    def get_actors_from_database(self) -> List[ActorAvro]:
        """
        Obtains a list of all active and suspended actors in the container.
        @return list of actors. Always non-null.
        """

    @abstractmethod
    def get_authorities(self) -> List[ActorAvro]:
        """
        Obtains a list of all active site authorities.
        @return list of actors. Always non-null.
        """

    @abstractmethod
    def get_brokers(self) -> List[ActorAvro]:
        """
        Obtains a list of all active brokers.
        @return list of actors. Always non-null.
        """

    @abstractmethod
    def get_controllers(self) -> List[ActorAvro]:
        """
        Obtains a list of all active controllers.
        @return list of actors. Always non-null.
        """

    @abstractmethod
    def get_proxies(self, *, protocol: str) -> List[ProxyAvro]:
        """
        Obtains a list of all proxies to actors using the specified protocol
        @param protocol protocol
        @return list of proxies. Always non-null.
        """

    @abstractmethod
    def get_broker_proxies(self, *, protocol: str) -> List[ProxyAvro]:
        """
        Obtains a list of all proxies to brokers using the specified protocol
        @param protocol protocol
        @return list of proxies. Always non-null.
        """

    @abstractmethod
    def get_authority_proxies(self, *, protocol: str) -> List[ProxyAvro]:
        """
        Obtains a list of all proxies to brokers using the specified protocol
        @param protocol protocol
        @return list of proxies. Always non-null.
        """

    @abstractmethod
    def get_management_object(self, *, key: ID) -> IComponent:
        """
        Obtains the specified management object
        @param key management object id
        @return management object proxy on success, null otherwise. The returned
                proxy will share the same credentials.
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabric.actor.core.apis.i_tick import ITick
    from fabric.actor.core.time.actor_clock import ActorClock


class IContainerClock:
    """
    IContainerClock describes the time-related container API.
    """

    @abstractmethod
    def get_actor_clock(self) -> ActorClock:
        """
        Returns an instance of the container's clock factory.
        @return container clock factory
        """

    @abstractmethod
    def tick(self):
        """
        Advances the container clock with one cycle (only if isManualClock() is true).
        """

    @abstractmethod
    def get_current_cycle(self):
        """
        Returns the current cycle.
        @return current clock cycle
        """

    @abstractmethod
    def stop(self):
        """
        Stops the clock.
        @throws Exception in case of error
        """

    @abstractmethod
    def register(self, *, tickable: ITick):
        """
        Registers an object with the ticker.
        @param tickable object to register
        """

    @abstractmethod
    def unregister(self, *, tickable: ITick):
        """
        Unregisters an object with the ticker.
        @param tickable object to register
        """

    @abstractmethod
    def is_manual_clock(self) -> bool:
        """
        Checks if the container clock advances manually.
        @return true if the container clock advances manually
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING, Dict

from fabric.actor.boot.inventory.neo4j_resource_pool_factory import Neo4jResourcePoolFactory
from fabric.actor.boot.inventory.resource_pool_factory import ResourcePoolFactory
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.core.pool_manager import PoolManagerError
from fabric.actor.core.plugins.config.configuration_mapping import ConfigurationMapping
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reflection_utils import ReflectionUtils
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType

if TYPE_CHECKING:
    from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
    from fabric.actor.boot.inventory.i_resource_pool_factory import IResourcePoolFactory
    from fabric.actor.core.plugins.substrate.authority_substrate import AuthoritySubstrate


class PoolCreator:
    """
    Responsible for setting up inventory pools on startup
    """
    def __init__(self, *, substrate: AuthoritySubstrate = None, pools: dict = None, neo4j_config: dict = None):
        self.substrate = substrate
        self.pools = pools
        self.neo4j_config = neo4j_config
        self.container = None
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

    def get_factory2(self):
        """
        Create Neo4j ResourcePool Factory instance
        """
        factory = Neo4jResourcePoolFactory()
        factory.set_substrate(substrate=self.substrate)
        return factory

    def get_factory(self, *, rd: ResourcePoolDescriptor) -> IResourcePoolFactory:
        """
        Create ResourcePool Factory instance
        @param rd resource pool descriptor
        """
        factory = None
        if rd.get_pool_factory_module() is None or rd.get_pool_factory_class() is None:
            factory = ResourcePoolFactory()
        else:
            try:
                factory = ReflectionUtils.create_instance_with_params(
                    module_name=rd.get_pool_factory_module(), class_name=rd.get_pool_factory_class())(self.neo4j_config)
            except Exception as e:
                raise PoolCreatorException("Could not instantiate class= {}.{} {}".format(
                    rd.get_pool_factory_module(), rd.get_pool_factory_class(), e))

        factory.set_substrate(substrate=self.substrate)
        factory.set_descriptor(descriptor=rd)
        return factory

    def process(self):
        """
        Create Pools
        """
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.container = GlobalsSingleton.get().get_container()
        for pool in self.pools.values():
            self.logger.debug("Creating resource pool {} of Actor {}".format(
                pool.get_resource_type_label(), self.substrate.get_actor().get_name()))

            factory = self.get_factory(rd=pool)
            pool = factory.get_descriptor()
            rd = ResourceData()
            rd.resource_properties = pool.save(properties=rd.resource_properties, prefix=None)

            rd.local_properties = ResourceData.merge_properties(from_props=pool.pool_properties,
                                                                to_props=rd.local_properties)

            create_pool_result = self.substrate.get_pool_manager().create_pool(slice_id=ID(),
                                                                               name=pool.get_resource_type_label(),
                                                                               rtype=pool.get_resource_type(),
                                                                               resource_data=rd)

            if create_pool_result.code != PoolManagerError.ErrorNone:
                raise PoolCreatorException("Could not create resource pool: {}. error={}".format(
                    pool.get_resource_type_label(), create_pool_result.code))

            self.register_handler(pool=pool)
            source = factory.create_source_reservation(slice_obj=create_pool_result.slice)

            try:
                self.logger.debug("Adding source reservation to database {}".format(source))
                self.logger.debug("Source reservation has resources of type {}"
                                  .format(source.get_resources().get_resources().__class__.__name__))
                self.logger.debug("Source reservation has delegation of type {}".format(
                    source.get_resources().get_resources().get_ticket().get_delegation().__class__.__name__))

                self.substrate.get_database().add_reservation(reservation=source)
            except Exception as e:
                raise PoolCreatorException("Could not add source reservation to database {}".format(e))

    def process_neo4j(self, substrate_file: str, actor_name: str) -> Dict:
        """
        Create Pools for Neo4j
        """
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.container = GlobalsSingleton.get().get_container()
        factory = self.get_factory2()

        rd = ResourceData()
        create_pool_result = self.substrate.get_pool_manager().create_pool(
            slice_id=ID(), name=actor_name, rtype=ResourceType(resource_type=Constants.property_aggregate_resource_model),
            resource_data=rd)

        if create_pool_result.code != PoolManagerError.ErrorNone:
            raise PoolCreatorException("Could not create resource pool: {}. error={}".format(
                actor_name, create_pool_result.code))

        self.logger.debug("Created aggregate manager resource slice# {}".format(create_pool_result.slice))

        arm_graph = None
        if create_pool_result.slice.get_graph_id() is not None:
            # load the graph from Neo4j database
            self.logger.debug("Reloading an existing graph for resource slice# {}".format(create_pool_result.slice))
            arm_graph = factory.get_arm_graph(graph_id=create_pool_result.slice.get_graph_id())
            create_pool_result.slice.set_graph(graph=arm_graph)
        else:
            arm_graph = factory.get_arm_graph_from_file(filename=substrate_file)
            create_pool_result.slice.set_graph(graph=arm_graph)
            self.substrate.get_pool_manager().update_pool(slice_obj=create_pool_result.slice)
            self.logger.debug("Created new graph for resource slice# {}".format(create_pool_result.slice))

        for pool in self.pools.values():
            self.logger.debug("Creating resource pool {} of Actor {}".format(
                pool.get_resource_type_label(), self.substrate.get_actor().get_name()))

            factory.set_descriptor(descriptor=pool)
            pool = factory.get_descriptor()

            self.register_handler(pool=pool)
            # TODO populate inventory and register controls
            # NEO4J: Discuss with Ilya

        return arm_graph.generate_adms()

    def register_handler(self, *, pool: ResourcePoolDescriptor):
        """
        Register Handlers
        @param pool Resource pool descriptor
        """
        handler_module = pool.get_handler_module()
        handler_class = pool.get_handler_class()

        if handler_class is None or handler_module is None:
            return

        config = self.substrate.get_config()
        config_map = ConfigurationMapping()
        config_map.set_key(key=str(pool.get_resource_type()))
        config_map.set_class_name(class_name=handler_class)
        config_map.set_module_name(module_name=handler_module)
        config_map.set_properties(properties=pool.get_handler_properties())

        config.add_config_mapping(mapping=config_map)


class PoolCreatorException(Exception):
    """
    Pool Creator Exception
    """
